# w.c.s. - web application for online forms
# Copyright (C) 2005-2012  Entr'ouvert
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import urlparse
import urllib
import sys

from quixote import get_request, get_publisher, get_response

from qommon import _
from qommon.form import *
from wcs.workflows import WorkflowStatusItem, register_item_class
from wcs.roles import get_user_roles, Role
from qommon.ident.idp import is_idp_managing_user_attributes
from qommon.misc import http_post_request, http_delete_request
from qommon.publisher import get_cfg, get_logger
from wcs.api_utils import sign_url, get_secret_and_orig, MissingSecret


def roles_ws_url(role_uuid, user_uuid):
    idps = get_cfg('idp', {})
    entity_id = idps.values()[0]['metadata_url']
    base_url = entity_id.split('idp/saml2/metadata')[0]
    url = urlparse.urljoin(base_url, '/api/roles/%s/members/%s/' % (urllib.quote(role_uuid),
                                                                     urllib.quote(user_uuid)))
    secret, orig = get_secret_and_orig(url)
    url += '?orig=%s' % orig
    return sign_url(url, secret)


class AddRoleWorkflowStatusItem(WorkflowStatusItem):
    description = N_('Add Role to User')
    key = 'add_role'

    role_id = None

    def get_parameters(self):
        return ('role_id',)

    def add_parameters_widgets(self, form, parameters, prefix='', formdef=None):
        if 'role_id' in parameters:
            form.add(SingleSelectWidget, '%srole_id' % prefix,
                     title=_('Role to Add'), value=str(self.role_id),
                     options=[(None, '----', None)] + get_user_roles())

    def role_id_export_to_xml(self, item, charset, include_id=False):
        self._role_export_to_xml('role_id', item, charset,
                include_id=include_id)

    def role_id_init_with_xml(self, elem, charset, include_id=False):
        self._role_init_with_xml('role_id', elem, charset,
                include_id=include_id)

    def perform(self, formdata):
        if not self.role_id:
            return
        self.role_id = str(self.role_id)
        if not formdata.user_id:
            # we can't work on anonymous forms
            return
        user = get_publisher().user_class.get(formdata.user_id)
        self.perform_local(user, formdata)
        if user.name_identifiers and is_idp_managing_user_attributes():
            self.perform_idp(user, formdata)

    def perform_local(self, user, formdata):
        if not user.roles:
            user.roles = []
        if not self.role_id in user.roles:
            user.roles.append(self.role_id)
        user.store()
        if get_request().user and get_request().user.id == user.id:
            # if we changed the currently logged in user, we update it with the
            # changes.
            get_request().user = user

    def perform_idp(self, user, formdata):
        role = Role.get(self.role_id)
        role_uuid = role.slug
        user_uuid = user.name_identifiers[0]
        try:
            url = roles_ws_url(role_uuid, user_uuid)
        except MissingSecret:
            get_publisher().notify_of_exception(sys.exc_info(), context='[ROLES]')
            return
        def after_job(job):
            response, status, data, auth_header = http_post_request(url)
            if status != 201:
                get_logger().error('failed to add role %r to user %r',
                                   role, user)
        get_response().add_after_job(
            str(N_('Adding role')),
            after_job)


register_item_class(AddRoleWorkflowStatusItem)


class RemoveRoleWorkflowStatusItem(WorkflowStatusItem):
    description = N_('Remove Role from User')
    key = 'remove_role'

    role_id = None

    def get_parameters(self):
        return ('role_id',)

    def add_parameters_widgets(self, form, parameters, prefix='', formdef=None):
        if 'role_id' in parameters:
            form.add(SingleSelectWidget, '%srole_id' % prefix,
                     title=_('Role to Remove'), value=self.role_id,
                     options=[(None, '----', None)] + get_user_roles())

    def perform(self, formdata):
        if not self.role_id:
            return
        self.role_id = str(self.role_id)
        if not formdata.user_id:
            # we can't work on anonymous forms
            return
        user = get_publisher().user_class.get(formdata.user_id)
        self.perform_local(user, formdata)
        if user.name_identifiers and is_idp_managing_user_attributes():
            self.perform_idp(user, formdata)

    def perform_local(self, user, formdata):
        if user.roles and self.role_id in user.roles:
            user.roles.remove(self.role_id)
            user.store()
            if get_request().user and get_request().user.id == user.id:
                # if we changed the currently logged in user, we update it
                # with the changes.
                get_request().user = user

    def perform_idp(self, user, formdata):
        role = Role.get(self.role_id)
        role_uuid = role.slug
        user_uuid = user.name_identifiers[0]
        try:
            url = roles_ws_url(role_uuid, user_uuid)
        except MissingSecret:
            get_publisher().notify_of_exception(sys.exc_info(), context='[ROLES]')
            return
        def after_job(job):
            response, status, data, auth_header = http_delete_request(url)
            if status != 200:
                get_logger().error('failed to remove role %r from user %r',
                                   role, user)
        get_response().add_after_job(
            str(N_('Removing role')),
            after_job)

register_item_class(RemoveRoleWorkflowStatusItem)
